/* vim: set sw=8: -*- Mode: C; tab-width: 8; indent-tabs-mode: t; c-basic-offset: 8 -*- */
/*
 * fn-numberstring.c:  NumberString function
 *
 * Author:
 *   Yukihiro Nakai (ynakai@redhat.com)
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include <gnumeric-config.h>
#include <glib/gi18n.h>
#include <gnumeric.h>
#include <func.h>

#include <parse-util.h>
#include <cell.h>
#include <sheet.h>
#include <format.h>
#include <workbook.h>
#include <expr.h>
#include <value.h>

#include "plugin.h"
#include "plugin-util.h"
#include "module-plugin-defs.h"

GNUMERIC_MODULE_PLUGIN_INFO_DECL;

/* format option enum */
enum {
	NUMFORMAT_NORMAL = 1,
	NUMFORMAT_TRADITIONAL,
	NUMFORMAT_SIMPLE,
	NUMFORMAT_NOEXIST /* For invalid option value detection */
};

static const guchar* numtable[] = {
	"",
	"",
	"",
	"",
	"",
	"",
	"ϻ",
	"",
	"Ȭ",
	"",
	NULL
};

static const guchar* oldnumtable[] = {
	"",
	"",
	"",
	"",
	"",
	"",
	"ϻ",
	"",
	"Ȭ",
	"",
	NULL,
};

static const guchar* ordertable[] = {
	"",
	"ɴ",
	"",
	"",
	"",
	"",
	NULL
};

static const guchar* oldordertable[] = {
	"",
	"ɴ",
	"",
	"",
	"",
	"",
	NULL
};

static const guchar*
get_numstring(int i, int fmtopt) {
	if( i < 0 || i > 9 )
		return NULL;
	if( fmtopt < 1 || fmtopt >= NUMFORMAT_NOEXIST )
		return NULL;

	if( fmtopt == NUMFORMAT_NORMAL || fmtopt == NUMFORMAT_SIMPLE )
		return numtable[i];

	if( fmtopt == NUMFORMAT_TRADITIONAL )
		return oldnumtable[i];

	return NULL;
}

static const guchar*
get_orderstring(int i, int fmtopt) {
	if( i < 0 || i > 5 )
		return NULL;
	if( fmtopt < 1 || fmtopt >= NUMFORMAT_NOEXIST )
		return NULL;

	if( fmtopt == NUMFORMAT_NORMAL || fmtopt == NUMFORMAT_SIMPLE )
		return ordertable[i];

	if( fmtopt == NUMFORMAT_TRADITIONAL )
		return oldordertable[i];

	return NULL;
}

/***************************************************************************/

static char const *help_numberstring = {
	N_("@FUNCTION=NUMBERSTRING\n"
	   "@SYNTAX=NUMBERSTRING(a, fmtopt)\n"

	   "@DESCRIPTION="
	   "NUMBERSTRING shows number string with Chinese number characters:"
	   "\n"
	   "@a is a normal number and @fmtopt is the output format option.\n"
	   "1 is the normal number format with order chars.\n"
           "2 is the traditional number format with order chars.\n"
           "3 is the normal number format without order chars.\n"
	   "\n"
	   "* If the values contain strings or empty cells, #NUM! is "
           "returned.\n"
	   "* If no logical values are provided, then the error #VALUE! "
	   "is returned.\n"
	   "* This function is Excel (Japanese and Korean version)"
           "compatible.\n")
};

static GnmValue *
gnumeric_numberstring (FunctionEvalInfo *ei, GnmValue **argv)
{
	int fmtopt;
	int strsize;
	guchar* tmpstr;
	gchar* result;
	GnmValue* rval;
	gint i;
	GnmDateConventions const *conv =
		workbook_date_conv(ei->pos->sheet->workbook);
	GnmFormat* format = style_format_new_XL("#", TRUE);

	fmtopt = value_get_as_int(argv[1]);

	if( fmtopt < 1 || fmtopt >= NUMFORMAT_NOEXIST )
		return value_new_error_NUM(ei->pos);

	if( !VALUE_IS_NUMBER(argv[0]) )
		return value_new_error_NUM(ei->pos);

	tmpstr = format_value(format, argv[0], NULL, -1, conv);

	style_format_unref(format);

	if( tmpstr == NULL )
		return value_new_error_NUM(ei->pos);

	result = g_strdup("");
	strsize = (int)strlen(tmpstr);
	for(i=0;i<strsize;i++) {
		gchar* tmptmp = g_strdup(result);
		switch(tmpstr[i]) {
		case '1':
			if( fmtopt == NUMFORMAT_NORMAL ) {
				int rest = strsize - i;
				if( rest == 1 || rest == 5 || rest == 9 || rest == 13 || rest >= 13 )
					result = g_strconcat(tmptmp, get_numstring(1, fmtopt), NULL);

			} else {
				result = g_strconcat(tmptmp, get_numstring(1, fmtopt), NULL);
			}
			break;
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			result = g_strconcat(tmptmp, get_numstring(tmpstr[i]-'1'+1, fmtopt), NULL);
			break;
		case '0':
			if( fmtopt == NUMFORMAT_SIMPLE )
				result = g_strconcat(tmptmp, get_numstring(0, fmtopt), NULL);
			break;
		default:
			break;
			
		}
		if( tmptmp != NULL )
			g_free(tmptmp);
		if( (fmtopt == NUMFORMAT_NORMAL || fmtopt == NUMFORMAT_TRADITIONAL ) && strsize - i <= 13) {
			tmptmp = g_strdup(result);
			switch(strsize - i) {
			case 2: /* 10 */
			case 6:
			case 10:
				if ( tmpstr[i] != '0' )
					result = g_strconcat(tmptmp, get_orderstring(0, fmtopt), NULL);
				break;
			case 3: /* 100 */
			case 7:
			case 11:
				if ( tmpstr[i] != '0' )
					result = g_strconcat(tmptmp, get_orderstring(1, fmtopt), NULL);
				break;
			case 4: /* 1000 */
			case 8:
			case 12:
				if ( tmpstr[i] != '0' )
					result = g_strconcat(tmptmp, get_orderstring(2, fmtopt), NULL);
				break;
			case 5: /* 10k */
				if( i-3 >= 0 && strncmp(tmpstr+i-3, "0000", 4) == 0 )
					break;
				result = g_strconcat(tmptmp, get_orderstring(3, fmtopt), NULL);
				break;
			case 9: /* 100 Million  - 100,000,000 */
				if( i-3 >= 0 && strncmp(tmpstr+i-3, "0000", 4) == 0 )
					break;
				result = g_strconcat(tmptmp, get_orderstring(4, fmtopt), NULL);
				break;
			case 13: /* 1 Trillion - 1,000,000,000,000 */
				result = g_strconcat(tmptmp, get_orderstring(5, fmtopt), NULL);
				break;
			default: /* No more upper order supported in Excel */
				break;
			}
			if( tmptmp != NULL )
				g_free(tmptmp);
		}
	}

	g_free(tmpstr);

	{
		GError* error;
		tmpstr = g_locale_to_utf8(result, -1, NULL, NULL, &error);
		g_free(result);
		result = tmpstr;
	}

        rval = value_new_string (result);
	g_free(result);

	return rval;
}


/***************************************************************************/

const GnmFuncDescriptor numberstring_functions[] = {
	{ "numberstring", "ff", N_("numstr,fmtopt,"), &help_numberstring,
	  gnumeric_numberstring, NULL, NULL, NULL, NULL,
	  GNM_FUNC_SIMPLE,
	  GNM_FUNC_IMPL_STATUS_COMPLETE, GNM_FUNC_TEST_STATUS_BASIC },
        {NULL}
};
